/****************************************************************************
 *  Copyright (C) 2018 RoboMaster.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 ***************************************************************************/
 
#include "pid.h"
#include "bsp_can.h"
extern moto_info_t motor_info[MOTOR_MAX_NUM];


const pid_struct_t M3508_pid_speed = {
		2, 0.005, 2, 16000, 16000
	};
const pid_struct_t M3508_pid_pos = {
		2, 0.005, 2, 16000, 16000
	};

const pid_struct_t GM6020_pid_speed = {
		40, 3, 0, 30000, 30000
	};
const pid_struct_t GM6020_pid_pos = {
		30, 4, 15, 10000, 10000
	};

/**
  * @brief  init pid parameter for one single type of control
  * @param  pid struct
    @param  parameters
  * @retval None
  */
		
void pid_param(pid_struct_t *pid,
              float kp,
              float ki,
              float kd,
              float i_max,
              float out_max)
{
  pid->kp      = kp;
  pid->ki      = ki;
  pid->kd      = kd;
  pid->i_max   = i_max;
  pid->out_max = out_max;
}


/**
  * @brief  init all pid parameters
  * @param  motor_pid_param_struct_t
    @param  pid struct
  * @retval None
  */
		
void pid_init(motor_pid_param_struct_t motor_list[1], pid_struct_t control_param[7])
{
	pid_param(&control_param[0], 2, 0.005, 2, 16000, 16000); // M3508-speed
	pid_param(&control_param[1], 2, 0.005, 2, 16000, 16000); // M3508-position
	pid_param(&control_param[2], 40, 3, 0, 30000, 30000);  // GM6020-speed
	pid_param(&control_param[3], 30, 4, 15, 10000, 10000);  // GM6020-position
	
	pid_motor_init(&motor_list[0], control_param[0], control_param[1]); // paramtres M3508
	pid_motor_init(&motor_list[1], control_param[2], control_param[3]); // paramtres GM6020
}


/**
  * @brief  init PID parameters for a motor (pos and speed)
  * @param  motor_pid_param_struct_t
    @param  pid struct
		@param  pid struct
  * @retval None
  */
		
void pid_motor_init(motor_pid_param_struct_t *pid_param,
						 	 pid_struct_t pos,
							 pid_struct_t speed)
{
	pid_param->position = pos;
	pid_param->speed = speed;
}

/**
  * @brief  pid calculation
  * @param  pid struct
    @param  reference value
    @param  feedback value
  * @retval calculation result
  */
float pid_calc(pid_struct_t *pid, float ref, float fdb)
{
  pid->ref = ref;
  pid->fdb = fdb;
  pid->err[1] = pid->err[0];
  pid->err[0] = pid->ref - pid->fdb;
  
  pid->p_out  = pid->kp * pid->err[0];
  pid->i_out += pid->ki * pid->err[0];
  pid->d_out  = pid->kd * pid->err[0] - pid->err[1];
  LIMIT_MIN_MAX(pid->i_out, -pid->i_max, pid->i_max);
  
  pid->output = pid->p_out + pid->i_out + pid->d_out;
  LIMIT_MIN_MAX(pid->output, -pid->out_max, pid->out_max);
  return pid->output;
}

void run_all_motors() /* send motor control message through can bus*/
{
	set_motor_voltage(0, 
                      motor_info[0].set_voltage, 
                      motor_info[1].set_voltage, 
                      motor_info[2].set_voltage, 
                      motor_info[3].set_voltage);
    
  set_motor_voltage(1, 
                      motor_info[4].set_voltage, 
                      motor_info[5].set_voltage, 
                      motor_info[6].set_voltage, 
                      0);
}


void run_motor_from_command(int id_motor, pid_struct_t pid_param, float target, float fdb)
{
	motor_info[id_motor].set_voltage = pid_calc(&pid_param, target, fdb);
}


void set_motor_position(int id_motor, motor_pid_param_struct_t motor, float target)
{
	run_motor_from_command(id_motor, motor.position, target, motor_info[id_motor].rotor_angle);
}


void set_motor_speed(int id_motor, motor_pid_param_struct_t motor, float target)
{
	run_motor_from_command(id_motor, motor.speed, target, motor_info[id_motor].rotor_speed);
}
